package provider

import (
	"context"
	"encoding/json"
	"fmt"
	"net/url"
	"sync"
	"time"

	"github.com/gorilla/websocket"
	"github.com/rs/zerolog"

	"github.com/umee-network/umee/price-feeder/oracle/types"
)

const (
	gateHost      = "ws.gate.io"
	gatePath      = "/v3"
	gatePingCheck = time.Second * 28 // should be < 30
)

var _ Provider = (*GateProvider)(nil)

type (
	// GateProvider defines an Oracle provider implemented by the Gate public
	// API.
	//
	// REF: https://www.gate.io/docs/websocket/index.html
	GateProvider struct {
		wsURL           url.URL
		wsClient        *websocket.Conn
		logger          zerolog.Logger
		reconnectTimer  *time.Ticker
		mtx             sync.RWMutex
		tickers         map[string]GateTicker         // Symbol => GateTicker
		candles         map[string][]GateCandle       // Symbol => GateCandle
		subscribedPairs map[string]types.CurrencyPair // Symbol => types.CurrencyPair
	}

	GateTicker struct {
		Last   string `json:"last"`       // Last traded price ex.: 43508.9
		Vol    string `json:"baseVolume"` // Trading volume ex.: 11159.87127845
		Symbol string `json:"symbol"`     // Symbol ex.: ATOM_UDST
	}

	GateCandle struct {
		Close     string // Closing price
		TimeStamp int64  // Unix timestamp
		Volume    string // Total candle volume
		Symbol    string // Total symbol
	}

	// GateTickerSubscriptionMsg Msg to subscribe all the tickers channels.
	GateTickerSubscriptionMsg struct {
		Method string   `json:"method"` // ticker.subscribe
		Params []string `json:"params"` // streams to subscribe ex.: BOT_USDT
		ID     uint16   `json:"id"`     // identify messages going back and forth
	}

	// GateCandleSubscriptionMsg Msg to subscribe to a candle channel.
	GateCandleSubscriptionMsg struct {
		Method string        `json:"method"` // ticker.subscribe
		Params []interface{} `json:"params"` // streams to subscribe ex.: ["BOT_USDT": 1800]
		ID     uint16        `json:"id"`     // identify messages going back and forth
	}

	// GateTickerResponse defines the response body for gate tickers.
	GateTickerResponse struct {
		Method string        `json:"method"`
		Params []interface{} `json:"params"`
	}

	// GateTickerResponse defines the response body for gate tickers.
	// The Params response is a 2D slice of multiple candles and their data.
	//
	// REF: https://www.gate.io/docs/websocket/index.html
	GateCandleResponse struct {
		Method string          `json:"method"`
		Params [][]interface{} `json:"params"`
	}

	// GateEvent defines the response body for gate subscription statuses.
	GateEvent struct {
		ID     int             `json:"id"`     // subscription id, ex.: 123
		Result GateEventResult `json:"result"` // event result body
	}
	// GateEventResult defines the Result body for the GateEvent response.
	GateEventResult struct {
		Status string `json:"status"` // ex. "successful"
	}
)

// NewGateProvider creates a new GateProvider.
func NewGateProvider(ctx context.Context, logger zerolog.Logger, pairs ...types.CurrencyPair) (*GateProvider, error) {
	wsURL := url.URL{
		Scheme: "wss",
		Host:   gateHost,
		Path:   gatePath,
	}

	wsConn, _, err := websocket.DefaultDialer.Dial(wsURL.String(), nil)
	if err != nil {
		return nil, fmt.Errorf("error connecting to Gate websocket: %w", err)
	}

	provider := &GateProvider{
		wsURL:           wsURL,
		wsClient:        wsConn,
		logger:          logger.With().Str("provider", "gate").Logger(),
		reconnectTimer:  time.NewTicker(gatePingCheck),
		tickers:         map[string]GateTicker{},
		candles:         map[string][]GateCandle{},
		subscribedPairs: map[string]types.CurrencyPair{},
	}
	provider.wsClient.SetPongHandler(provider.pongHandler)

	if err := provider.SubscribeCurrencyPairs(pairs...); err != nil {
		return nil, err
	}

	go provider.handleReceivedTickers(ctx)

	return provider, nil
}

// GetTickerPrices returns the tickerPrices based on the saved map.
func (p *GateProvider) GetTickerPrices(pairs ...types.CurrencyPair) (map[string]TickerPrice, error) {
	tickerPrices := make(map[string]TickerPrice, len(pairs))

	for _, currencyPair := range pairs {
		price, err := p.getTickerPrice(currencyPair)
		if err != nil {
			return nil, err
		}

		tickerPrices[currencyPair.String()] = price
	}

	return tickerPrices, nil
}

// GetCandlePrices returns the candlePrices based on the saved map
func (p *GateProvider) GetCandlePrices(pairs ...types.CurrencyPair) (map[string][]CandlePrice, error) {
	candlePrices := make(map[string][]CandlePrice, len(pairs))

	for _, currencyPair := range pairs {
		gp := currencyPairToGatePair(currencyPair)
		price, err := p.getCandlePrices(gp)
		if err != nil {
			return nil, err
		}

		candlePrices[currencyPair.String()] = price
	}

	return candlePrices, nil
}

func (p *GateProvider) getCandlePrices(key string) ([]CandlePrice, error) {
	p.mtx.RLock()
	defer p.mtx.RUnlock()

	candles, ok := p.candles[key]
	if !ok {
		return []CandlePrice{}, fmt.Errorf("failed to get candle prices for %s", key)
	}

	candleList := []CandlePrice{}
	for _, candle := range candles {
		cp, err := candle.toCandlePrice()
		if err != nil {
			return []CandlePrice{}, err
		}

		candleList = append(candleList, cp)
	}

	return candleList, nil
}

// SubscribeCurrencyPairs subscribe to ticker and candle channels for all pairs.
func (p *GateProvider) SubscribeCurrencyPairs(cps ...types.CurrencyPair) error {
	if err := p.subscribeTickers(cps...); err != nil {
		return err
	}
	if err := p.subscribeCandles(cps...); err != nil {
		return err
	}
	p.setSubscribedPairs(cps...)
	return nil
}

// subscribeTickers subscribes to the ticker channels for all pairs at once.
func (p *GateProvider) subscribeTickers(cps ...types.CurrencyPair) error {
	topics := []string{}

	for _, cp := range cps {
		topics = append(topics, currencyPairToGatePair(cp))
	}

	tickerMsg := newGateTickerSubscription(topics...)
	if err := p.subscribeTickerPairs(tickerMsg); err != nil {
		return err
	}

	return nil
}

// subscribeCandles subscribes to the candle channels for all pairs one-by-one.
// The gate API currently only supports subscribing to one kline market at a time.
//
// REF: https://www.gate.io/docs/websocket/index.html
func (p *GateProvider) subscribeCandles(cps ...types.CurrencyPair) error {
	gatePairs := make([]string, len(cps))

	iterator := 0
	for _, cp := range cps {
		gatePairs[iterator] = currencyPairToGatePair(cp)
		iterator++
	}

	for _, pair := range gatePairs {
		msg := newGateCandleSubscription(pair)
		if err := p.subscribeCandlePair(msg); err != nil {
			return err
		}
	}

	return nil
}

func (p *GateProvider) subscribedPairsToSlice() []types.CurrencyPair {
	p.mtx.RLock()
	defer p.mtx.RUnlock()

	return mapPairsToSlice(p.subscribedPairs)
}

func (p *GateProvider) getTickerPrice(cp types.CurrencyPair) (TickerPrice, error) {
	p.mtx.RLock()
	defer p.mtx.RUnlock()

	gp := currencyPairToGatePair(cp)
	if tickerPair, ok := p.tickers[gp]; ok {
		return tickerPair.toTickerPrice()
	} else {
		return TickerPrice{}, fmt.Errorf("gate provider failed to get ticker price for %s", gp)
	}
}

func (p *GateProvider) handleReceivedTickers(ctx context.Context) {
	for {
		select {
		case <-ctx.Done():
			return
		case <-time.After(defaultReadNewWSMessage):
			messageType, bz, err := p.wsClient.ReadMessage()
			if err != nil {
				// if some error occurs continue to try to read the next message.
				p.logger.Err(err).Msg("could not read message")
				if err := p.ping(); err != nil {
					p.logger.Err(err).Msg("could not send ping")
				}
				continue
			}

			if len(bz) == 0 {
				continue
			}

			p.resetReconnectTimer()
			p.messageReceived(messageType, bz)

		case <-p.reconnectTimer.C: // reset by the pongHandler.
			if err := p.reconnect(); err != nil {
				p.logger.Err(err).Msg("error reconnecting")
			}
		}
	}
}

func (p *GateProvider) messageReceived(messageType int, bz []byte) {
	if messageType != websocket.TextMessage {
		return
	}

	var gateEvent GateEvent
	if err := json.Unmarshal(bz, &gateEvent); err != nil {
		p.logger.Debug().Msg("received a message that is not an event")
	} else {
		switch gateEvent.Result.Status {
		case "success":
			return
		case "":
			break
		default:
			p.reconnect()
		}
	}

	if err := p.messageReceivedTickerPrice(bz); err != nil {
		// msg is not a ticker, it will try to marshal to candle message.
		p.logger.Debug().Err(err).Msg("unable to unmarshal ticker")
	} else {
		return
	}
	if err := p.messageReceivedCandle(bz); err != nil {
		p.logger.Debug().Err(err).Msg("unable to unmarshal candle")
	}
}

// messageReceivedTickerPrice handles the ticker price msg.
// The provider response is a slice with different types at each index.
//
// REF: https://www.gate.io/docs/websocket/index.html
func (p *GateProvider) messageReceivedTickerPrice(bz []byte) error {
	var tickerMessage GateTickerResponse
	if err := json.Unmarshal(bz, &tickerMessage); err != nil {
		return err
	}

	if tickerMessage.Method != "ticker.update" {
		return fmt.Errorf("message is not a ticker update")
	}

	tickerBz, err := json.Marshal(tickerMessage.Params[1])
	if err != nil {
		p.logger.Err(err).Msg("could not marshal ticker message")
		return err
	}

	var gateTicker GateTicker
	if err := json.Unmarshal(tickerBz, &gateTicker); err != nil {
		p.logger.Err(err).Msg("could not unmarshal ticker message")
		return err
	}

	symbol, ok := tickerMessage.Params[0].(string)
	if !ok {
		return fmt.Errorf("symbol should be a string")
	}
	gateTicker.Symbol = symbol

	p.setTickerPair(gateTicker)
	return nil
}

// UnmarshalParams is a helper function which unmarshals the 2d slice of interfaces
// from a GateCandleResponse into the GateCandle.
func (gc *GateCandle) UnmarshalParams(params [][]interface{}) error {
	var tmp []interface{}

	if len(params) == 0 {
		return fmt.Errorf("no candles in response")
	}

	// use the most recent candle
	tmp = params[len(params)-1]
	if len(tmp) != 8 {
		return fmt.Errorf("wrong number of fields in candle")
	}

	time := int64(tmp[0].(float64))
	if time == 0 {
		return fmt.Errorf("time field must be a float")
	}
	gc.TimeStamp = time

	close, ok := tmp[1].(string)
	if !ok {
		return fmt.Errorf("close field must be a string")
	}
	gc.Close = close

	volume, ok := tmp[5].(string)
	if !ok {
		return fmt.Errorf("volume field must be a string")
	}
	gc.Volume = volume

	symbol, ok := tmp[7].(string)
	if !ok {
		return fmt.Errorf("symbol field must be a string")
	}
	gc.Symbol = symbol

	return nil
}

// messageReceivedCandle handles the candle price msg.
// The provider response is a slice with different types at each index.
//
// REF: https://www.gate.io/docs/websocket/index.html
func (p *GateProvider) messageReceivedCandle(bz []byte) error {
	var candleMessage GateCandleResponse
	if err := json.Unmarshal(bz, &candleMessage); err != nil {
		return err
	}

	if candleMessage.Method != "kline.update" {
		return fmt.Errorf("message is not a kline update")
	}

	var gateCandle GateCandle
	if err := gateCandle.UnmarshalParams(candleMessage.Params); err != nil {
		return err
	}

	p.setCandlePair(gateCandle)
	return nil
}

func (p *GateProvider) setTickerPair(ticker GateTicker) {
	p.mtx.Lock()
	defer p.mtx.Unlock()
	p.tickers[ticker.Symbol] = ticker
}

func (p *GateProvider) setCandlePair(candle GateCandle) {
	p.mtx.Lock()
	defer p.mtx.Unlock()
	// convert gate timestamp seconds -> milliseconds
	candle.TimeStamp = candle.TimeStamp * int64(time.Second/time.Millisecond)
	staleTime := PastUnixTime(providerCandlePeriod)
	candleList := []GateCandle{}

	candleList = append(candleList, candle)
	for _, c := range p.candles[candle.Symbol] {
		if staleTime < c.TimeStamp {
			candleList = append(candleList, c)
		}
	}
	p.candles[candle.Symbol] = candleList
}

// subscribeTickerPairs write the subscription msg to the provider.
func (p *GateProvider) subscribeTickerPairs(msg GateTickerSubscriptionMsg) error {
	return p.wsClient.WriteJSON(msg)
}

// subscribeCandlePair write the subscription msg to the provider.
func (p *GateProvider) subscribeCandlePair(msg GateCandleSubscriptionMsg) error {
	return p.wsClient.WriteJSON(msg)
}

// setSubscribedPairs sets N currency pairs to the map of subscribed pairs.
func (p *GateProvider) setSubscribedPairs(cps ...types.CurrencyPair) {
	p.mtx.Lock()
	defer p.mtx.Unlock()

	for _, cp := range cps {
		p.subscribedPairs[cp.String()] = cp
	}
}

func (p *GateProvider) resetReconnectTimer() {
	p.reconnectTimer.Reset(gatePingCheck)
}

// reconnect closes the last WS connection and creates a new one. If there’s a
// network problem, the system will automatically disable the connection. The
// connection will break automatically if the subscription is not established or
// data has not been pushed for more than 30 seconds. To keep the connection stable:
// 1. Set a timer of N seconds whenever a response message is received, where N is
// less than 30.
// 2. If the timer is triggered, which means that no new message is received within
// N seconds, send the String 'ping'.
// 3. Expect a 'pong' as a response. If the response message is not received within
// N seconds, please raise an error or reconnect.
func (p *GateProvider) reconnect() error {
	p.wsClient.Close()

	p.logger.Debug().Msg("reconnecting websocket")
	wsConn, _, err := websocket.DefaultDialer.Dial(p.wsURL.String(), nil)
	if err != nil {
		return fmt.Errorf("error reconnecting to Gate websocket: %w", err)
	}
	wsConn.SetPongHandler(p.pongHandler)
	p.wsClient = wsConn

	currencyPairs := p.subscribedPairsToSlice()
	return p.SubscribeCurrencyPairs(currencyPairs...)
}

// ping to check websocket connection.
func (p *GateProvider) ping() error {
	return p.wsClient.WriteMessage(websocket.PingMessage, ping)
}

func (p *GateProvider) pongHandler(appData string) error {
	p.resetReconnectTimer()
	return nil
}

func (ticker GateTicker) toTickerPrice() (TickerPrice, error) {
	return newTickerPrice("Gate", ticker.Symbol, ticker.Last, ticker.Vol)
}

func (candle GateCandle) toCandlePrice() (CandlePrice, error) {
	return newCandlePrice(
		"Gate",
		candle.Symbol,
		candle.Close,
		candle.Volume,
		candle.TimeStamp,
	)
}

// currencyPairToGatePair returns the expected pair for Gate
// ex.: "ATOM_USDT".
func currencyPairToGatePair(pair types.CurrencyPair) string {
	return pair.Base + "_" + pair.Quote
}

// newGateTickerSubscription returns a new subscription topic for tickers.
func newGateTickerSubscription(cp ...string) GateTickerSubscriptionMsg {
	return GateTickerSubscriptionMsg{
		Method: "ticker.subscribe",
		Params: cp,
		ID:     1,
	}
}

// newGateCandleSubscription returns a new subscription topic for candles.
func newGateCandleSubscription(gatePair string) GateCandleSubscriptionMsg {
	params := []interface{}{
		gatePair, // currency pair ex. "ATOM_USDT"
		60,       // time interval in seconds
	}
	return GateCandleSubscriptionMsg{
		Method: "kline.subscribe",
		Params: params,
		ID:     2,
	}
}
